//
// ecRuleOp.h
//
// $Id: ecRuleOp.h,v 1.3 2000/09/20 05:36:15 mcoletti Exp $
//
// Contains genetic operators specialized for rule learning using EAs

#ifndef EC_RULEOP_H
#define EC_RULEOP_H

#include <algorithm>

#ifndef EC_GENOP_H
#include <eclib/ecGenOp.h>
#endif

#ifndef EC_RANDOM_H
#include <eclib/ecRandom.h>
#endif

typedef set<int> CrossoverPoints;

/**

 class ecNPointRuleBoundaryCrossoverOperator

 This operator will work on variable length genomes which are composed
 of fixed length rules (Michalewicz refers to these rules or subcomponents
 as complexes in his book).  Crossovers will occur only on rule boundaries.

 Individuals are required to have the following accessors:

    int minSize()
    int maxSize()
    int ruleSize()

 Note: The minSize affects this crossover operator's ability to function.
       The minSize should always equal the number of crossover points
       plus 1.

 The resulting individuals are guaranteed to be the within the size
 limitations.

 The one-point version of this operator is equivalent to the crossover
 operator Stephen Smith used in his LS-1 system for his dissertation at
 the University of Pittsburgh in 1980.  He also implemented a version
 which operated on non-rule boundaries.

 The one-point version of this operator was also used by Wu, Schultz and
 Agah in the original MAV experiments.

*/
template<class Individual>
class ecNPointRuleBoundaryCrossoverOperator
    : public ecBinaryGeneticOperator<Individual>
{
public:

    ecNPointRuleBoundaryCrossoverOperator(int   numPoints = 1, // xover points
                                          float pCross = 1.0,  // prob of xover
                                          int   maxAttempts = 10,
                                          ecGeneticOperator<Individual>* provider = NULL)
        : ecBinaryGeneticOperator<Individual>(provider),
          numPoints_(numPoints),
          pCross_(pCross),
          maxAttempts_(maxAttempts)
    {}

    virtual ~ecNPointRuleBoundaryCrossoverOperator() {}

    virtual void apply(Individual& ind1, Individual& ind2)
    {
        // Check pCross
        if ( drand() > pCross_ )
            return;

        // Check lengths
        // XXX Asserts are probably not the best way to do error checking.

        assert(ind1.ruleSize() == ind2.ruleSize());
        assert((int)ind1.size() / ind1.ruleSize() > numPoints());
        assert((int)ind2.size() / ind2.ruleSize() > numPoints());

        // Find crossover points.  Crossover points are stored as a
        // set of unique indices into the individuals.  Each index
        // indicates the gene before which the crossover will occur.
        // In other words if the index is 1, a crossover will be made
        // between gene 0 and gene 1.
        CrossoverPoints  xpoints1, xpoints2;
        int size1, size2;
        int attempts = 0;

        do
        {
            if ( attempts > maxAttempts() )
                return;   // Individuals are unchanged.
            attempts++;

            while ((int)xpoints1.size() < numPoints() )
            {
                int rule_boundary = irand((ind1.size() / ind1.ruleSize())-1)+1;
                xpoints1.insert( rule_boundary * ind1.ruleSize() );
            }
            xpoints1.insert( ind1.size() );  // A null terminator of sorts.

            while ((int)xpoints2.size() < numPoints() )
            {
                int rule_boundary = irand((ind2.size() / ind2.ruleSize())-1)+1;
                xpoints2.insert( rule_boundary * ind2.ruleSize() );
            }
            xpoints2.insert( ind2.size() );  // A null terminator of sorts.

            calculateOffspringSizes (xpoints1, xpoints2, size1, size2);
        } while (size1 < ind1.minSize() || size1 > ind1.maxSize() ||
                 size2 < ind2.minSize() || size2 > ind2.maxSize() );

        // Perform the crossovers

        // Temporary variables to store the new genomes.
        // XXX I'm concerned that this approach may create a conflict
        // XXX with Mark's mechanism for storing allele bounds.
        vector<typename Individual::value_type> genome1, genome2;

        typename Individual::iterator i1 = ind1.begin();
        typename Individual::iterator i2 = ind2.begin();
        int g1 = 0;  // keep track of the indices of the genes
        int g2 = 0;

        bool swap_data = false;
        typename CrossoverPoints::iterator x1, x2;
        for (x1 = xpoints1.begin(), x2 = xpoints2.begin();
             x1 != xpoints1.end();
             x1++, x2++)
        {
            while (g1 < *x1)
            {
                if (swap_data)
                    genome2.push_back(*i1);
                else
                    genome1.push_back(*i1);

                ++i1;
                ++g1;
            }

            while (g2 < *x2)
            {
                if (swap_data)
                    genome1.push_back(*i2);
                else
                    genome2.push_back(*i2);

                ++i2;
                ++g2;
            }

            swap_data = !swap_data;
        }

        // copy the new genomes back into the original individuals
        ind1.clear();
        ind2.clear();

        vector<typename Individual::value_type>::iterator iter;
        for (iter = genome1.begin(); iter != genome1.end(); iter++)
            ind1.push_back(*iter);

        for (iter = genome2.begin(); iter != genome2.end(); iter++)
            ind2.push_back(*iter);
    }

    int numPoints() const  { return numPoints_; }
    int numPoints(int n)   { return numPoints_ = n; }

    float pCross() const    { return pcross_; }
    float pCross( float s ) { return pcross_ = s; }

    int maxAttempts() const  { return maxAttempts_; }
    int maxAttempts(int n)   { return maxAttempts_ = n; }

private:

    // number of crossover points.
    int numPoints_;

    // probability that the crossover operation will be performed.
    float pCross_;

    // maximum number of attempts to be made at finding appropriate
    // crossover points before giving up.
    int maxAttempts_;

    void calculateOffspringSizes(const CrossoverPoints &  cross1,
                                 const CrossoverPoints &  cross2,
                                 int &  size1,
                                 int &  size2)
    {
        size1 = 0;
        size2 = 0;

        int oldx1 = 0;
        int oldx2 = 0;
        bool swap_data = false;
        CrossoverPoints::iterator x1, x2;

        for (x1 = cross1.begin(), x2 = cross2.begin();
             x1 != cross1.end();
             x1++, x2++)
        {
            if (swap_data)
            {
                size1 = size1 + *x2 - oldx2;
                size2 = size2 + *x1 - oldx1;
            }
            else
            {
                size1 = size1 + *x1 - oldx1;
                size2 = size2 + *x2 - oldx2;
            }
            
            oldx1 = *x1;
            oldx2 = *x2;
            swap_data = !swap_data;
        }
    }


}; // class ecNPointCrossoverOperator


#endif

