//
// ecRuleOp_t.cpp
//
// $Id: ecRuleOp_t.cpp,v 1.4 2000/10/18 19:13:04 mcoletti Exp $
//


#include <algorithm>
#include <iterator>
#include <functional>
#include <numeric>
#include <iostream>
#include <cassert>
#include <deque>


using namespace std;


#include <eclib/ecIndividual.h>
#include <eclib/ecPopulation.h>
#include <eclib/ecRandom.h>
#include <eclib/ecSelector.h>
#include <eclib/ecGenOp.h>
#include <eclib/ecBirth.h>
#include <eclib/ecReproStrategy.h>
#include <eclib/ecSurvive.h>
#include <eclib/ecSurvivalStrategy.h>
#include <eclib/ecHalt.h>
#include <eclib/ecStatistics.h>
#include <eclib/ecEA.h>

#include <contrib/rules/ecRuleOp.h>


// A test individual class for the 1-max landscape
//
// Used to exercise rule learning oriented tests.
//
class BinaryRuleIndividual : public ecIndividual< bool, float, vector<bool> >
{
public:

    BinaryRuleIndividual(int initialSize = 0,  // These defaults don't make
                         int minSize = 0,      // sense, but we must have a
                         int maxSize = 0,      // default ctor.
                         int ruleSize = 0)
        : ecIndividual< bool, float, vector<bool> >( initialSize ),
          minSize_(minSize),
          maxSize_(maxSize),
          ruleSize_(ruleSize)
    {}

    void randomize()
    {
        // randomize the genes to values in [0,1]
        for ( iterator i = begin(); i != end(); i++ )
        {
            *i = irand(2);
        }
    }

    // The 1-max problem provides selection pressure for larger rule sets.
    float evaluate() const
    {
        return accumulate( begin(), end(), 0.0 );
    }

    int minSize() {return minSize_;} const
    int minSize(int min_size) {return minSize_ = min_size;}

    int maxSize() {return maxSize_;} const
    int maxSize(int max_size) {return maxSize_ = max_size;}

    int ruleSize() {return ruleSize_;} const
    int ruleSize(int rule_size) {return ruleSize_ = rule_size;}

protected:

    void print_( ostream & os ) const
    {
        os << "( ";

        ostream_iterator<float> ostr_itr( os, " " );
        copy( begin(), end(), ostr_itr );

        os << "), "
           << fitness() 
           << "\n";
    }

private:

    int  minSize_;
    int  maxSize_;
    int  ruleSize_;

}; // BinaryRuleIndividual



// Add some syntactic sugar so we don't have to type long names all the time

// Abstract classes
typedef ecPopulation<BinaryRuleIndividual>                Population;
typedef ecSelector<Population>                            Selector;
typedef ecGeneticOperator<BinaryRuleIndividual>           GeneticOperator;
typedef ecBirthOperator<Population,Selector>              BirthOperator;
typedef ecReproductionStrategy<Population,BirthOperator>  ReproductionStrategy;
typedef ecSurvivalOperator<Population,Selector>           SurvivalOp;
typedef ecSurvivalStrategy<Population,SurvivalOp>         SurvivalStrategy;
typedef ecStatistics<Population>                          Statistics;
typedef ecHalt<Population>                                Halt;


// Concrete classes
typedef ecSequentialSelector<Population>                     SequentialSelector;
typedef deque<GeneticOperator*>                                Pipeline;
typedef ecPipelineBirthOperator<Population,Selector,Pipeline>  PipelineBOP;
typedef ecBinaryMutationBirthOperator<Population,Selector>     BinMutBOP;
typedef ecTruncationSurvival<Population,Selector>              TruncSurvivalOp;
typedef ecMuPlusLambdaSurvivalStrategy<Population,SurvivalOp>  MuPlusLambda;
typedef ecGenerationHalt<Population>                           StopAtGeneration;


int
main( int argc, char** argv )
{
    int  mu = 20;      // Number of parents
    int  lambda = 30;  // Number of children
    int  max_generations = 30;

    int  ruleSize = 3;
    int  minSize = ruleSize * 3;    // 3 rules - minimum for 2-point crossover
    int  maxSize = ruleSize * 100;
    int  initSize = ruleSize * 5;

    Population parents;

    // let's add some guys to the parents
    for ( int i = 0; i < mu; i++ )
    {
        parents.push_back( BinaryRuleIndividual(initSize, minSize,
                                                maxSize, ruleSize) );
    }

    ostream_iterator<BinaryRuleIndividual> ostr_itr( cout, "\n" );

    // give them initial random values
    randomize( parents.begin(), parents.end() );

    cout << "parents:\n";
    copy( parents.begin(), parents.end(), ostr_itr );

    // the next three steps necessary to make a create function to be
    // passed into the EA may indicate a design problem -- a lot of
    // hand waving here


   // first build the reproduction strategy components, and then the
   // reproductive strategy itself

   SequentialSelector birth_sequential_selector;

   ecBitFlipMutationOperator<BinaryRuleIndividual>             mutate(0.01);
   ecNPointRuleBoundaryCrossoverOperator<BinaryRuleIndividual> cross(2, 1.0);

   Pipeline pipeline;
   pipeline.push_back(&mutate);
   pipeline.push_back(&cross);

   PipelineBOP pipeline_bop( birth_sequential_selector, pipeline );
   pipeline_bop.init( parents );


   ReproductionStrategy   reproducer( pipeline_bop );


   // next, build the surival strategy components followed by the
   // survival strategy

   SequentialSelector survival_sequential_selector;

   TruncSurvivalOp  truncationSurvival( survival_sequential_selector );

   MuPlusLambda     muPlusLambda( truncationSurvival );


   StopAtGeneration stopAtGeneration( max_generations );

   Statistics  statistics;

   // invoke a simple ES style algorithm
   ecEA< Population,
         StopAtGeneration,
         ReproductionStrategy,
         SurvivalStrategy,
         Statistics
       >
      ( parents,
        lambda,
        stopAtGeneration,
        reproducer,
        muPlusLambda,
        statistics
      );


   cout << "parents after " << max_generations << " generations:\n";

   copy( parents.begin(), parents.end(), ostr_itr );

   exit( 0 );
}


