//
// ecBirth.h
//
// $Id: ecBirth.h,v 1.12 2001/05/01 18:02:28 mcoletti Exp $
//
//


#ifndef EC_BIRTH_H
#define EC_BIRTH_H

#include <algorithm>
#include <queue>



#ifndef EC_RANDOM_H
#include <eclib/ecRandom.h>
#endif


#ifndef EC_OPERATOR_H
#include <eclib/ecOperator.h>
#endif


#ifndef EC_GENOP_H
#include <eclib/ecGenOp.h>
#endif



/// Abstract base class for all birth operators.
template< class Population, class SelectionOperator >
class ecBirthOperator : public ecSingleOperator<Population>
{
   public:

      ecBirthOperator( SelectionOperator & so )
         : selection_operator_( so )
      {}

      virtual ~ecBirthOperator() {}


      /// Returns a new child.
      virtual typename Population::value_type operator()( ) = 0;

      /// Accessor for selection operator
      SelectionOperator const & selector() const
      { return selection_operator_; }

      /// Accessor for selection operator
      SelectionOperator & selector()
      { return selection_operator_; }


      ///Initialize the birth operator
      /** Initialize the birth operator, which entails always
        initializing the selector, and then doing any birth operator
        specific initialization, if any. 
      */
      void init( Population & first )
      {
         selector().init( first );

         init_( first );
      }

      /// The second population is ignored by birth operators.
      /** 

          \verbatim

       XXX this is a kludge -- part of the problem with Strategy design
       XXX patterns is that occasionally some parameters are superfluous
       XXX depending on context; this is one of those instances.  Need
       XXX to explore ways around this.

       \endverbatim

       \sa init( Population & first )
      */
      void init( Population & first, Population & second )
      {
         init_( first, second );
      }

      ///Do any post-processing.  
      /** Do any post-processing.  The child birth operator can define
        the semantics for this, if any.
      */
      void cleanup( Population & first )
      {
         cleanup_( first );
      }

      ///Do any post-processing. 
      /**
         \sa init(Population & first, Population & second)
       */
      void cleanup( Population & first, Population & second )
      {
         cleanup_( first, second );
      }


   protected:

      //@{
      /** 
          These are counterparts to init() and cleanup() that can be
          optionally over-ridden by child classes.
      */
      virtual void init_( Population & first )
      {
         // NOP
      }

      virtual void cleanup_( Population & first )
      {
         // NOP
      }

      virtual void init_( Population & first, Population & second )
      {
         // NOP
      }

      virtual void cleanup_( Population & first, Population & second )
      {
         // NOP
      }
      //@}

   private:

      ///reference to operator used to select parents
      SelectionOperator & selection_operator_;

}; // class ecBirthOperator




/// This is a simple clone birth operator.
/** 
    This birth operator just returns the selected individual with no
    modification.  This may be useful for experiments where one is
    testing selection and/or survival characteristics.  
*/
template<class Population, class SelectionOperator>
class ecCloneBirthOperator
   : public ecBirthOperator<Population,SelectionOperator>
{
   public:

      ecCloneBirthOperator( SelectionOperator & so )
         : ecBirthOperator<Population,SelectionOperator>( so )
      {}

      virtual ~ecCloneBirthOperator() {}

      /// just return the selected individual without modification
      virtual typename Population::value_type operator()( ) 
      {
         return selector()();
      }

}; // class ecCloneBirthOperator





/// An gaussian mutation operator -- genes better be real!
/**

 \note 

 What about clamping values to within a given range? NO, then it
 wouldn't support ES.  ES has no notion of explicit boundaries.

 \todo XXX we can probably generalize this further; the float argument is  kinda a kludge

*/
template<class Gene>
struct ecGaussianMutateGene
{
      ecGaussianMutateGene( float scale, float probability = 0.5 )
         : scale_( scale ), probability_( probability )
      {}

      /**
         Given a gene, perturb it with probability ``probability''
      */
      Gene & operator()( Gene & gene )
      {
         if ( frand() < probability_ )
         {
            return gene = gene + gauss() * scale_;
         }

         return gene;           // i.e., a NOP, so return unmodified value
      }

      /// how much to perturb
      float scale_;

      /// likelihood of pertubing value
      float probability_;

}; // class ecGaussianMutateGene




/// A simple gaussian mutation operator
template<class Population, class SelectionOperator>
class ecGaussianMutationBirthOperator 
   : public ecBirthOperator<Population,SelectionOperator>
{
   public:

      ecGaussianMutationBirthOperator( SelectionOperator & so,
                                     float scale, 
                                     float probability )
         : ecBirthOperator<Population,SelectionOperator>( so ),
           scale_( scale ),
           probability_( probability )
      {}

      virtual ~ecGaussianMutationBirthOperator() {}

      ///
      virtual typename Population::value_type operator()( ) 
      {
         // this of course presumes that the child's genes are
         // real-valued; this should generate a compiler error if it
         // isn't

         // copy over a parent's genes to the new child; note that the
         // child was created by the caller but has no value as yet
         typename Population::value_type child = selector()();

         // perturb each child component (i.e., gene)
         std::for_each( child.begin(), 
                        child.end(), 
                        ecGaussianMutateGene<typename Population::value_type::value_type>( scale_, probability_ )  );

         return child;
      }

      ///
      float scale() const    { return scale_; }
      ///
      float scale( float s ) { return scale_ = s; }

      ///
      float probability() const    { return probability_; }
      ///
      float probability( float s ) { return probability_ = s; }


   private:

      /// how wide should the gaussian distribution be?
      float scale_;
      
      /// what are the chances that we mutate a given gene?
      float probability_;
      
}; // class ecGaussianMutationBirthOperator




// a uniform crossover helper function
/**
 This will create two children from two parents.  One will be
 returned; the other returned on the subsequent invocation.

 \note

 This presumes that the parents are of equal length.
 
 \verbatim
 XXX Presumably there's some way we can enforce subtype policy
 XXX similar to the one that the STL uses with regards to iterator
 XXX type safety. E.g., we want to insure that variable length
 XXX individuals uses a different crossover mechanism.
 \endverbatim

 \todo ripped out probability bias ... need to talk to Jeff about this

*/
template<class Individual>
class
uniform_crossover__
{
   public:


      Individual operator()( Individual const & first,
                             Individual const & second )
      {
                                // Return the second child created from the last
                                // invocation, if one exists.

         if ( ! cache_.empty() )
         {

            Individual tmp_child = cache_.front();

            cache_.pop();

            return tmp_child;
         }


         Individual first_child;
         Individual second_child;

         typename Individual::const_iterator i = first.begin();
         typename Individual::const_iterator j = second.begin();

         for ( ;
               i != first.end() && j != second.end();
               i++, j++ )
         {
            if ( drand() < 0.5 )
            {
               first_child.push_back( *i );
               second_child.push_back( *j );
            }
            else
            {
               first_child.push_back( *j );
               second_child.push_back( *i );
            }

         }

         // arbitrarily stuff the first kid into the cache and return
         // the second; it'll be returned the next time this is called

         cache_.push( first_child );

         return second_child;

      } // uniform_crossover__


      /// insure that we're not holding on to old kids from previous generations
      void cleanup()
      {
         // clear out the cache for new work
         while ( cache_.size() )
         {
            cache_.pop();
         }
      }

   private:

      /// private cache of kids
      /**
         The uniform crossover, like any two parent crossover
         operator, generates two children.  However, the interface
         only returns a single child.  So, when children are created,
         one is cached, and the other returned.  The cached child is
         returned on the next call.  The cache is flushed so that that
         the cycle repeats.

         \todo XXX This makes sense for 0.5 ratios, but what about biased selections?
      */
      std::queue<Individual> cache_;

}; // functor uniform_crossover__





/// A simple uniform crossover operator
/**
 Not by itself useful, but good for a base class for adding functionality.
*/
template<class Population, class SelectionOperator>
class ecUniformCrossoverBirthOperator
   : public ecBirthOperator<Population,SelectionOperator>
{
   public:

      ecUniformCrossoverBirthOperator( SelectionOperator & so )
         : ecBirthOperator<Population,SelectionOperator>( so )
      {}

      virtual ~ecUniformCrossoverBirthOperator() {}

 
      /// return a new child
      virtual typename Population::value_type operator()( ) 
      {
         // Select two parents from the parent population
         typename Population::value_type & first_parent  = 
            selector()();
         typename Population::value_type & second_parent = 
            selector()();

         return crossover_operator_( first_parent, second_parent );
      }

      /// insure that the crossover operator's cache is cleared
      void cleanup_( Population & p )
      {
         crossover_operator_.cleanup( );
      }


   private:

      uniform_crossover__<typename Population::value_type> crossover_operator_;

}; // class ecUniformCrossoverBirthOperator






/// A simple uniform crossover and mutation operator
/**
 Inherits from ecGaussianMutationBirthOperator to get at the scale and
 probabiliy state.
*/
template<class Population, class SelectionOperator>
class ecUniformMutationBirthOperator :
   public ecGaussianMutationBirthOperator<Population,SelectionOperator>
{
   public:

      ecUniformMutationBirthOperator( SelectionOperator & so,
                                    float scale,
                                    float probability )
         :  ecGaussianMutationBirthOperator<Population,SelectionOperator>
              ( so, scale, probability )
      {}

      virtual ~ecUniformMutationBirthOperator() {}

      /// return a new child
      virtual typename Population::value_type operator()( ) 
      {
         // First create the child via crossover and then apply
         // gaussian mutation to it.

         // Select two parents from the parent population
         typename Population::value_type & first_parent  = 
            selector()();
         typename Population::value_type & second_parent = 
            selector()();

         typename Population::value_type child =
              crossover_operator_( first_parent, second_parent );
         
         // perturb each child component (i.e., gene)
         std::for_each( child.begin(), 
                        child.end(), 
                        ecGaussianMutateGene<typename Population::value_type::value_type>( scale(), 
                                                                                         probability() )  );

         return child;

      }

      /// insure that the crossover operator's cache is cleared
      void cleanup_( Population & p )
      {
         crossover_operator_.cleanup();
      } // cleanup_

   private:

      uniform_crossover__<typename Population::value_type> crossover_operator_;

}; // class ecUniformMutationBirthOperator




/** A very simple per-gene binary mutation help functor.  Note that
 this presumes use of ISO C++ STL vector<bool> specialization, or
 something that supports the notion of a ::flip() operator for
 toggling a bit's state.
*/
template<class Gene>
struct ecBinaryMutateGene
{
      ecBinaryMutateGene( float probability ) 
         : probability_( probability ) // better be in [0.0,1.0]
      {}

      ///
     Gene & operator()( Gene & gene )
      {
         if ( frand() < probability_ )
         {
            gene = ! gene;      // toggle bit
            return gene;
         }
         return gene;           // i.e., a NOP, so return unmodified value
      }

      float probability_;       // do we, or do we not apply the mutation?

}; // class ecBinaryMutateGene




/// A simple binary mutation operator
template<class Population, class SelectionOperator>
class ecBinaryMutationBirthOperator
   : public ecBirthOperator<Population,SelectionOperator>
{
   public:

      ecBinaryMutationBirthOperator( SelectionOperator & so, 
                                   float              probability )
         : ecBirthOperator<Population,SelectionOperator>( so ),
           probability_( probability )
      {}

      virtual ~ecBinaryMutationBirthOperator() {}

      ///
      virtual typename Population::value_type operator()( ) 
      {
         // this of course presumes that the child's genes are
         // real-valued; this should generate a compiler error if it
         // isn't

         // copy over a parent's genes to the new child; note that the
         // child was created by the caller but has no value as yet
         typename Population::value_type child = selector()();

         // perturb each child component (i.e., gene)
         for( typename Population::value_type::iterator i = 
                 child.begin();
              i != child.end();
              i++ )
         {
            if ( drand() < probability_ ) *i = ! (*i);
         }
      
         return child;
      }

      ///
      float probability() const    { return probability_; }

      ///
      float probability( float s ) { return probability_ = s; }


   private:

      /// what are the chances that we mutate a given gene? 
      float probability_;
      

}; // class ecBinaryMutationBirthOperator





/// A simple uniform crossover and binary mutation operator
/**
 Inherits from ecBinaryMutationBirthOperator to get at the scale and
 probabiliy state.
*/
template<class Population, class SelectionOperator>
class ecUniformBinaryMutationBirthOperator :
   public ecBinaryMutationBirthOperator<Population,SelectionOperator>
{
   public:

      ecUniformBinaryMutationBirthOperator( SelectionOperator & so,
                                          float probability )
         :  ecBinaryMutationBirthOperator<Population,SelectionOperator>
              ( so, probability )
      {}

      virtual ~ecUniformBinaryMutationBirthOperator() {}

      ///
      virtual typename Population::value_type operator()( ) 
      {
         // First create the child via crossover and then apply
         // binary mutation to it.

         // Select two parents from the parent population
         typename Population::value_type & first_parent  = 
            selector()();
         typename Population::value_type & second_parent = 
            selector()();

         typename Population::value_type child =
              crossover_operator_( first_parent, second_parent );
         
         // perturb each child component (i.e., gene)

         for( typename Population::value_type::iterator i = 
                 child.begin();
              i != child.end();
              i++ )
         {
            if ( drand() < probability() ) *i = ! (*i);
         }
      
         return child;

      }

      /// insure that the crossover operator's cache is cleared
      void
      cleanup_( Population & p )
      {
         crossover_operator_.cleanup();
      } // cleanup_

   private:
      uniform_crossover__<typename Population::value_type> crossover_operator_;

}; // class ecUniformBinaryMutationBirthOperator



template< class Population, class SelectionOperator, class Pipeline >
class ecPipelineBirthOperator
    : public ecBirthOperator<Population, SelectionOperator>
{
   public:

      ecPipelineBirthOperator( SelectionOperator & so,
                               Pipeline & pipeline )
          : ecBirthOperator<Population,SelectionOperator>(so),
            copy_operator_(so),
            pipeline_(pipeline),
            pipeline_end_(&copy_operator_)
      {
          // Connect up the pipeline
          typename Pipeline::iterator iter;

          pipeline_end_ = &copy_operator_;   // The first segment
          for (iter = pipeline.begin(); iter != pipeline.end(); ++iter)
          {
              (*iter)->provider(pipeline_end_);
              pipeline_end_ = *iter;
          }
      }

      virtual ~ecPipelineBirthOperator() {}

      // Returns a new child.
      virtual typename Population::value_type operator()( )
      {
          typename Population::value_type  result;
          typename Population::value_type* result_ptr;

          result_ptr = pipeline_end_->get();
          result = *result_ptr;
          delete result_ptr;    // The pipeline allocates new children

          return result;
      }

      /// insure that the operators in the pipeline clear their caches
      void
      cleanup_( Population & p )
      {
          // cleanup() is called recursively back to the beginning of
          // the pipeline.
          pipeline_end_->cleanup();
      }


   private:

      ecCopyOperator<Population> copy_operator_;
      // reference to implicit pipeline operator used to copy parents
      // Note: ecCopyOperator is templated on Population instead of
      //       Individual because it needs to be connected to a selector.

      Pipeline & pipeline_;

      ecGeneticOperator<typename Population::value_type>* pipeline_end_;

}; // class ecPipelineBirthOperator


#endif


