//
// ecEA.h
//
// $Id: ecEA.h,v 1.10 2000/10/18 19:13:04 mcoletti Exp $
//

#ifndef EC_EA_H
#define EC_EA_H

/**
 Given a halting function, a creation function, and a survive
 function, evolve the given population until the halting function
 returns true.

 
 \param parents 
 A container pre-sized to the desired number of parents; moreover,
 they should be initialized prior to calling ecEA().

 \param lambda
 The number of children to be produced every iteration.

 \param halt
 An object that controls when to stop evolving.

 \param reproductionStrategy
 An object that manages how children are created.

 \param survivalStrategy
 An object that controls which individuals will be used in the next iteration.

 \param statistics
 An object for collecting evolution statistics.

*/
template<class Population, 
         class Halt, 
         class ReproductionStrategy, 
         class SurvivalStrategy, 
         class Statistics>
void
ecEA( Population            & parents,
      unsigned int            lambda,
      Halt const            & halt,
      ReproductionStrategy  & reproductionStrategy,
      SurvivalStrategy      & survivalStrategy,
      Statistics            & statistics )
{
   // create a container of lambda children
   Population children( lambda ); 


   // create a scratch container for survivors; the survival strategy
   // needs to have the survivor container properly sized
   Population survivors( parents.size() ); 


   // seed the statistics with snapshot of parents before evolution begins

   statistics( parents );


   // Continue evolving the population until the halting function
   // returns true; note that it generally needs access to population
   // state information to determine when to stop.

   while( ! halt(parents) )
   {
      // Create a new population from the old.

      reproductionStrategy( parents, children );
                                // XXX further diagnostic info would be good

      // Now determine who the survivors are. Note that some Survive
      // implementations may choose to ignore the parents when
      // creating the new set of parents as is the case for
      // (mu,lambda) survival strategies.

      survivalStrategy.cull( parents, children, survivors );


      // We use a swap instead of copy as swapping can be constant
      // time vs. the linear time for copying.  (Though it obviously
      // depends on the population's container type.)  That is,
      // swapping may just be a manner of exchanging pointers to
      // internal datastructures, as opposed to doing an element-wise
      // copying.

      parents.swap( survivors );


      // now update any statistics

      statistics( parents, children );

   }

}; // ecEA


#endif
























