//
// ecGenOp.h
//
// $Id: ecGenOp.h,v 1.5 2000/10/19 18:11:24 jbassett Exp $
//

#ifndef EC_GENOP_H
#define EC_GENOP_H

#include <set>
#include <vector>
#include <queue>
#include <algorithm>

#ifndef EC_OPERATOR_H
#include <eclib/ecOperator.h>
#endif

#ifndef EC_SELECTOR_H
#include <eclib/ecSelector.h>
#endif

#ifndef EC_RANDOM_H
#include <eclib/ecRandom.h>
#endif


/**

 class ecGeneticOperator

 Abstract base class for all genetic operators.
 Genetic operators work only on individuals, and have no concept of a
 population.  They get their individuals from a Provider, and are
 Providers themselves.

*/
template< class Individual >
class ecGeneticOperator
{
public:

    ecGeneticOperator( ecGeneticOperator* provider = NULL )
        : provider_( provider )
    {}

    virtual ~ecGeneticOperator() {}

    virtual Individual* get() = 0;

    ecGeneticOperator* provider() { return provider_; }

      /** Attaches this operator to a provider, potentially unattaching
          it from another provider in the process. */
    void provider(ecGeneticOperator* provider) { provider_ = provider; }

      /** Attaches an entire segment of a pipeline to a provider.
     Will recursively call attach() on each ecGeneticOperator in the
     pipeline until a NULL is found. */
    void attach(ecGeneticOperator* new_provider)
    {
        if (provider() == NULL)
            provider(new_provider);
        else
            provider()->attach(new_provider);
    }

    ///Do any post-processing.  
    /** Do any post-processing.  The child birth operator can define
        the semantics for this, if any.
    */
    void cleanup(void)
    {
        cleanup_();

        // Call cleanup() recursively through the pipeline back to the 
        // beginning.
        if (provider() != NULL)
            provider()->cleanup();
    }

private:

    // pointer to operator used to provide a new individual
    ecGeneticOperator* provider_;

      /** 
          These is a counterpart to cleanup() that can be
          optionally over-ridden by child classes.
      */
      virtual void cleanup_( void )
      {
         // NOP
      }

}; // class ecGeneticOperator



/**

 class ecCopyOperator

 Makes an exact copy of the individual received from a selector.
 Allocate new memory.

 This genetic operator does not get it's data the provider.  Instead
 it gets it from a selector.  It is the only way to connect a pipeline
 of genetic operators to a selector.

*/
template<class Population>
class ecCopyOperator
    : public ecGeneticOperator<typename Population::value_type>
{
public:
    
    ecCopyOperator( ecSelector<Population>& selector )
        : ecGeneticOperator<typename Population::value_type>( NULL ),
          selector_(selector)
    {}

    virtual ~ecCopyOperator() {}

    virtual typename Population::value_type* get()
    {
        typename Population::value_type& ind = selector() ();

        // XXX Should we use clone here?
        typename Population::value_type* myCopy 
            = new typename Population::value_type(ind);

        return myCopy;
    }

    ecSelector<Population>& selector() {return selector_;}
    ecSelector<Population>& selector(ecSelector<Population> selector)
        {return selector_ = selector;}

/*
    ecGeneticOperator<typename Population::value_type>* provider()
    {
        return NULL;
    }

    ecGeneticOperator<typename Population::value_type>*
        provider(ecGeneticOperator<typename Population::value_type>* provider)
    {
        return NULL;
    }

    void attach(ecGeneticOperator* new_provider) {}
*/

private:
    ecSelector<Population>&  selector_;

}; // class ecCopyOperator



/**

 class ecCachingGeneticOperator

 Contains a mechanism for caching partial solutions.  For example, a
 crossover operator may create two offspring, but only one can be
 returned through the get() method.  We cache the second offspring
 until the next time get() is called.

*/
template< class Individual >
class ecCachingGeneticOperator
    : public ecGeneticOperator < Individual >
{
public:

    ecCachingGeneticOperator( ecGeneticOperator<Individual>* provider )
        : ecGeneticOperator<Individual>(provider)
    {}

    virtual ~ecCachingGeneticOperator() {}

    virtual Individual* get()
    {
        Individual* result = NULL;
        if ( !cache_.empty() )
        {
            result = cache_.front();
            cache_.pop();

            return result;
        }

        if (provider() == NULL)
            return NULL;

        apply();

        // Make sure a child was actually created.
        if ( cache_.empty() )
            result = NULL;
        else
        {
            result = cache_.front();
            cache_.pop();
        }

        return result;
    }

    virtual void addToCache(Individual* new_ind)
    {
        cache_.push(new_ind);
    }

    // This contains the guts of the genetic operator
    virtual void apply() = 0;

private:

    queue<Individual*>  cache_;

    virtual void cleanup_( void )
    {
        // Empty the cache
        while (!cache_.empty())
            cache_.pop();
    }

}; // class ecCachingGeneticOperator



/**

 class ecUnaryGeneticOperator

*/
template< class Individual >
class ecUnaryGeneticOperator
    : public ecGeneticOperator < Individual >
{
public:

    ecUnaryGeneticOperator( ecGeneticOperator<Individual>* provider)
        : ecGeneticOperator<Individual>(provider)
    {}

    virtual ~ecUnaryGeneticOperator() {}

    virtual Individual* get()
    {
        if (provider() == NULL)
            return NULL;
        
        Individual* ind_ptr = provider()->get();
        if (ind_ptr == NULL)
            return NULL;
        
        Individual& ind = *ind_ptr;
        
        // Call virtual apply method
        apply(ind);
        
        return ind_ptr;
    }

    /// This contains the guts of the genetic operator
    virtual void apply(Individual& ind) = 0;

}; // class ecUnaryGeneticOperator



/**

 class ecBinaryGeneticOperator

 Binary in the sense that it take two parameters, not in the sense that
 it deals with 1s and 0s.

*/
template< class Individual >
class ecBinaryGeneticOperator
    : public ecCachingGeneticOperator < Individual >
{
public:

    ecBinaryGeneticOperator( ecGeneticOperator<Individual>* provider)
        : ecCachingGeneticOperator<Individual>(provider)
    {}

    virtual ~ecBinaryGeneticOperator() {}

    virtual void apply()
    {
        Individual* ind1_ptr = provider()->get();
        if (ind1_ptr == NULL)
            return;

        Individual* ind2_ptr = provider()->get();
        if (ind2_ptr == NULL)
            return;
        
        Individual& ind1 = *ind1_ptr;
        Individual& ind2 = *ind2_ptr;
        
        // Call virtual apply method
        apply(ind1, ind2);

        addToCache(ind1_ptr);
        addToCache(ind2_ptr);
    }

    // This contains the guts of the genetic operator
    virtual void apply(Individual& ind1, Individual& ind2) = 0;

}; // class ecBinaryGeneticOperator



/**

 class ecBitFlipMutationOperator

 A simple mutation operator for binary genes.

 As long as your genes implement the ! operater, this genetic operator
 can be used on them.

 This operator will modify the individual it gets from the Provider.
 It will not make a copy.

*/
template<class Individual>
class ecBitFlipMutationOperator
    : public ecUnaryGeneticOperator<Individual>
{
   public:

      ecBitFlipMutationOperator(float              pMutate,
                                ecGeneticOperator<Individual>* provider = NULL)
         : ecUnaryGeneticOperator<Individual>(provider),
           pMutate_(pMutate)
      {}

      virtual ~ecBitFlipMutationOperator() {}

      virtual void apply(Individual& ind)
      {
         // perturb each child component (i.e., gene)
         for( typename Individual::iterator i = ind.begin();
              i != ind.end();
              i++ )
         {
            if ( drand() < pMutate() )
                *i = ! (*i);
         }
      }

      float pMutate() const    { return pMutate_; }
      float pMutate( float s ) { return pMutate_ = s; }


   private:

      /// what are the chances that we mutate a given gene?
      float pMutate_;


}; // class ecBitFlipMutationOperator



/**

 class ecUniformCrossoverOperator

 This operator requires that the 2 individuals be the same length.

 \todo XXX I should add a probability of crossover.

*/
template<class Individual>
class ecUniformCrossoverOperator
    : public ecBinaryGeneticOperator<Individual>
{
public:

    ecUniformCrossoverOperator(float              pSwapGene = 0.5,
                               ecGeneticOperator<Individual>* provider = NULL)
        : ecBinaryGeneticOperator<Individual>(provider),
        pSwapGene_(pSwapGene)
    {}

    virtual ~ecUniformCrossoverOperator() {}

    virtual void apply(Individual& ind1, Individual& ind2)
    {
        // Check lengths
        assert(ind1.size() == ind2.size());

        // swap genes if
        typename Individual::iterator i1, i2;
        for (i1 = ind1.begin(), i2 = ind2.begin();
             i1 != ind1.end();
             i1++, i2++ )
        {
            if ( drand() < pSwapGene() )
                swap (*i1, *i2);
        }
    }

    float pSwapGene() const    { return pSwapGene_; }
    float pSwapGene( float s ) { return pSwapGene_ = s; }


private:

    // probability that any given gene will be crossed.
    float pSwapGene_;

}; // class ecUniformCrossoverOperator



/**

 class ecNPointCrossoverOperator

 This operator requires that the 2 individuals be the same length.

*/
template<class Individual>
class ecNPointCrossoverOperator
    : public ecBinaryGeneticOperator<Individual>
{
public:

    ecNPointCrossoverOperator(int         numPoints = 1,  // xover points
                              float       pCross = 1.0,   // prob of xover
                              ecGeneticOperator<Individual>* provider = NULL)
        : ecBinaryGeneticOperator<Individual>(provider),
          numPoints_(numPoints),
          pCross_(pCross)
    {}

    virtual ~ecNPointCrossoverOperator() {}

    virtual void apply(Individual& ind1, Individual& ind2)
    {
        // Check pCross
        if ( drand() > pCross() )
            return;

        // Check lengths
        // XXX Asserts are probably not the best way to do error checking.
        assert(ind1.size() == ind2.size());
        assert(ind1.size() > numPoints());  // make sure we have enough genes.

        // Find crossover points.  Crossover points are stored as a
        // set of unique indices into the individuals.  Each index
        // indicates the gene before which the crossover will occur.
        // In other words if the index is 1, a crossover will be made
        // between gene 0 and gene 1.
        // XXX This has the potential to be very inefficient.  If the number
        // XXX of crossover points is very close to the number of potential
        // XXX crossover points, we could spend a lot of time in this loop.

        set<int>  xpoint_indices;

        while (xpoint_indices.size() < numPoints() )
            xpoint_indices.insert( irand(ind1.size()-1) + 1 );

        typedef typename Individual::iterator  IndIter;

        // Perform the crossovers
        bool swap_data = false;
        typename set<int>::iterator xiter = xpoint_indices.begin();
        IndIter i1, i2;
        int curr_gene;
        for (i1 = ind1.begin(), i2 = ind2.begin(), curr_gene = 0;
             i1 != ind1.end();
             i1++, i2++, curr_gene++ )
        {
            if (xiter != xpoint_indices.end() && curr_gene == *(xiter) )
            {
                swap_data = !swap_data;
                xiter++;
            }

            if ( swap_data )
                swap (*i1, *i2);
        }

/*
        // This section would replace the "perform crossover" section above.
        // I did all this in the hopes of making it more effecient and
        // easier to understand.  It is more effecient.  I don't think
        // its easier to understand though.

        // Create two lists of iterators.  Each iterator represents
        // a crossover point.  These iterators are stored in pairs where
        // the first iterator is a start point and the second is an
        // end point.
        vector<IndIter>  xpoints1;  // crossovers in ind1
        vector<IndIter>  xpoints2;  //     "      "  ind2
        IndIter p1 = ind1.begin();
        IndIter p2 = ind2.begin();

        typename set<int>::iterator xiter;
        for (xiter = xpoint_indices.begin();
             xiter != xpoint_indices.end();
             xiter++)
        {
            // XXX Need to advance by a relative offset.
            advance(p1, *xiter);
            advance(p2, *xiter);
            xpoints1.push_back(p1);
            xpoints2.push_back(p2);
        }

        // Deal with odd numbers of crossover points.  Make sure there
        // are always pairs of crossover points.
        if (numPoints() % 2 == 1)
        {
            xpoints1.push_back( ind1.end() );
            xpoints2.push_back( ind2.end() );
        }

        // Perform the crossovers
        typename vector<IndIter>::iterator start1, start2, end1;
        for (start1 = xpoints1.begin(), start2 = xpoints2.begin();
             start1 != xpoints1.end();
             advance(start1, 2), advance(start2, 2) ) // advance pair
        {
            end1 = start1;
            ++end1;
            IndIter i1, i2;
            for (i1 = *start1, i2 = *start2;
                 i1 != *end1;
                 i1++, i2++ )
            {
                swap (*i1, *i2);
            }
        }
*/
    }

    int numPoints() const  { return numPoints_; }
    int numPoints(int n)   { return numPoints_ = n; }

    float pCross() const    { return pCross_; }
    float pCross( float s ) { return pCross_ = s; }


private:

    /// number of crossover points.
    int numPoints_;

    /// probability that the crossover operation will be performed.
    float pCross_;

}; // class ecNPointCrossoverOperator


#endif

