//
// ecIndividual.h
//
// $Id: ecIndividual.h,v 1.11 2000/10/18 19:13:04 mcoletti Exp $

#ifndef EC_INDIVIDUAL_H
#define EC_INDIVIDUAL_H

#include <deque>
#include <iostream>

/**
 A generic individual.  It has a representation, a fitness type, and
 a genome, which is a container of values of the representation type.

 Like ecPopulation, this is an adaptor pattern for a genome, which
 is a container of data (genes)
// XXX Consider Genomes that are not of simple types ... say a smart
// XXX gene that knows its own bounds

*/
#ifndef __STL_LIMITED_DEFAULT_TEMPLATES
template <class Representation, class Fitness, class Genome = deque<Representation> >
#else   
template<class Representation, class Fitness, class Genome>
#endif
class ecIndividual
{
   public:

      typedef typename Genome::value_type value_type;
      typedef typename Genome::size_type size_type;
      typedef typename Genome::reference reference;
      typedef typename Genome::const_reference const_reference;
      typedef typename Genome::iterator iterator;
      typedef typename Genome::const_iterator const_iterator;
      typedef Fitness fitness_type;

      ecIndividual( size_type n, Representation const & init_value )
          : genome_( n, init_value ), needsEvaluation_( true )
      {}

      ecIndividual( size_type n )
          : genome_( n ), needsEvaluation_( true )
      {}

      ecIndividual( )
          : needsEvaluation_( true )
      {}

      virtual ~ecIndividual() {}



      iterator begin()
      {
          needsEvaluation_ = true;
          return genome_.begin();
      }

      const_iterator begin() const { return genome_.begin(); }


      iterator end()
      {
          needsEvaluation_ = true;
          return genome_.end();
      }

      const_iterator end() const { return genome_.end(); }

      ///
      reference operator[](size_type __n)
      {
          needsEvaluation_ = true;
          return *(begin() + __n);
      }

      ///
      const_reference operator[](size_type __n) const { return *(begin() + __n); }

      ///
      void push_back( value_type const v )
      { 
         genome_.push_back( v );
         needsEvaluation_ = true;
      }

      ///
      size_type size() const { return genome_.size(); }

      /// clear the individual's genes
      /**

         \note

         This means that the genes are out-right deleted, not set to
         some null value. 
      */
      void clear()
      {
          genome_.clear();
          needsEvaluation_ = true;
      }

      /// each individual must have a way to set itself to random values
      /**
         \todo XXX could this just be an external algorithm instead?
      */
      virtual void randomize() = 0;


      /// calculates the individual's fitness
      /** 

          \note 

         the fitness is a template parameter as this could be
         anything, even a non-numeric value

      */
      virtual Fitness evaluate() const = 0;


      /** returns the individual's fitness; the fitness is a template
         parameter as this could be anything even a non-numeric value
      */
      virtual Fitness fitness() const
      {
          if (needsEvaluation_)
          {
              fitness_ = evaluate();
              needsEvaluation_ = false;
          }

          return fitness_;
      }


      ///
      bool operator<( ecIndividual const & rhs ) const
      {
         return compare_( rhs );
      }

   protected:

      /** print function defined by each ecIndividual subclass; used
       by ostream operator<<
      */
      virtual void print_( ostream & os ) const = 0;


      /** a simple, default comparison function used by operator<(),
       which in turn will be used by STL less<> dependent operators
      */
      virtual bool compare_( ecIndividual const & i ) const
      {
         return fitness() < i.fitness();
      }

   private:

      /// all the gene values
      Genome genome_;

      /// the fitness value of this individual
      /**
         \note

         mutable because sometimes we need to non-chalantly set this in const members

       */
      mutable Fitness fitness_;

      /** indicates whether the current fitness_ is valid, or needs to be
          recalculated
      */
      mutable bool needsEvaluation_;

      friend ostream& operator<< <Representation,Fitness,Genome>( ostream & os, 
                                  ecIndividual<Representation,Fitness,Genome> const & );

}; // class ecIndividual





template<class Representation, class Fitness, class Genome>
inline
ostream&
operator<<( ostream & os, 
            ecIndividual<Representation,Fitness,Genome> const & i )
{
   i.print_( os );

   return os;
} // operator<<( ostream & os, ecIndividual const & i )


// XXX Should we define a RC smart pointer for possible use for
// XXX heterogenous individual containers?


#endif






