//
// ecSelector.h
//
// $Id: ecSelector.h,v 1.8 2000/11/21 06:59:13 mcoletti Exp $
//

#ifndef EC_SELECTOR_H
#define EC_SELECTOR_H

#include <list>
#include <algorithm>


#ifndef EC_OPERATOR_H
#include <eclib/ecOperator.h>
#endif



/// Abstract base class for all selectors
template<class Population>
class ecSelector : public ecDoubleOperator<Population>
{
   public:

      ecSelector() 
         : mode_( single_pop ),
           first_pop_( 0x0 ),
           second_pop_( 0x0 )
      {}

      virtual ~ecSelector() {}


      /// Select and return an individual
      /** 
         An individual is selected from a population and the passed in
         individual is set to its value.
         XXX is this counterintuitive?
      */
      virtual typename Population::value_type & get() = 0;


      /// Select and return an individual
      typename Population::value_type & operator()( )
      {
         return get();
      }

      /// Initialize a selector with the given population
      /**
         This sets up the selector to choose individuals from the
         given population.  It also sets it into "single population"
         mode. */
      virtual void init( Population & first )
      {
         // we're just working with a single population
         mode_ = single_pop;

         // bind to the given population
         first_pop_ = & first;

         // let the sub-class do any initialization unique to it
         init_( first );
      }


      /// Initialize a selector with the given populations
      /**
         This sets up the selector to choose individuals from the
         given populations.  It also sets it into "double population"
         mode; i.e., it gets individuals from \b two populations. */
      virtual void init( Population & first, Population & second )
      {
         // we're working with two populations
         mode_ = double_pop;

         // bind to the given populations
         first_pop_  = & first;
         second_pop_ = & second;

         init_( first, second );
      }


// THESE SHOULD BE DEFINED IN SUB-CLASSES
//
//        virtual void cleanup( Population & first )
//        {
//        }
//
//        virtual void cleanup( Population & first, Population & second )
//        {
//        }

   protected:

      /** type definition for whether the selector is using just the
       first population, or both populations
      */
      typedef enum { single_pop, double_pop } mode_type;

      /// return whether we're selecting from one or two populations 
      mode_type mode() const { return mode_; }

      /// 
      Population const & first()  const { return *first_pop_; }
      Population const & second() const { return *second_pop_; }

      ///
      Population &       first()        { return *first_pop_; }
      Population &       second()       { return *second_pop_; }

      /// the subclass version of init()
      virtual void init_( Population & first ) = 0;
      virtual void init_( Population & first, Population & second ) = 0;

   private:

      /// state for selecting from one or two populations
      mode_type  mode_;

      /// pointer to first population we're selecting from
      Population * first_pop_;

      /// pointer to second population, if any, we're selecting from
      Population * second_pop_;
      
}; // class ecSelector





/// Return an arbitrary individual from the population.
template<class Population>
class ecUniformSelector : public ecSelector<Population>
{
   public:

      /// select an invididual
      /**
         \fn virtual typename Population::value_type & get()

         \todo XXX verify that I got the population selection bias correct

         Select and return an individual from one or two populations.
         If the selector is in "single population mode," then the
         individual is selected from the first() population.
         Otherwise the selector is in "two population mode," which
         means the selected individual can come from either the
         first() or second() population.  

         \note

         In "two population mode," a population is first arbitrarily
         selected biased by population size.

       */
      virtual typename Population::value_type & get()
      {
         if ( ecSelector<Population>::single_pop == mode() )
         {
            return first()[ irand( first().size() )  ];
         }
         else
         {
            // Arbitrarily pick a population and select from it; but
            // be fair and bias the population choice by population
            // size.  'f' and 'p' are in [0,1]; 'p' is calculated as
            // the ratio of the first population to the second; 'f'
            // uses that ratio to determine which population to choose
            // from.

            float f = drand();
            float p = static_cast<float>(first().size()) / 
                   static_cast<float>(first().size() + second().size()) ;

            if ( f < p )
            {
               return first()[ irand( first().size() )  ];
            }
            else
            {
               return second()[ irand( second().size() )  ];
            }
         }
      }

      /// do any post-processing cleanup
      virtual void cleanup( Population & first )
      {
         // NOP
      }
      virtual void cleanup( Population & first, Population & second )
      {
         // NOP
      }

   private:

      /// allow child classes to do any initialization
      void init_( Population & first )
      {
         // NOP
      }
      void init_( Population & first, Population & second )
      {
         // NOP
      }

}; // class UniformSelector




/// Returns each individual in the population one at a time.
/**
 If there are TWO populations, it'll enumerate the first population,
 and then the second.  At the end of the second population, the
 iterator is reset to the beginning of the first population.
*/
template<class Population>
class ecSequentialSelector : public ecSelector<Population>
{
   public:

      ecSequentialSelector()
         : done_first_( false )
      {}

      /// select and return an individual in a linear sequence
      /**
         Sequentially select and return individuals from one or more
         populations.  If in "single population mode," return an
         individual from the first() population; otherwise return an
         individual from either the first() or second() population. */
      virtual typename Population::value_type & get()
      {
         // The single population case is simple.  We just return the
         // current individual and increment to the next; we wrap
         // around to the first after the last individual.
         if ( ecSelector<Population>::single_pop == mode() )
         {
            if ( first().end() == current_ )
            {
               current_ = first().begin();
            }
         }
         // Unfortunately things are a little more complicated for the
         // two population model.  The premise is this: iterate
         // through the first population, and then the second.  When
         // done with the second population, wrap back around to the
         // first.  We use ``done_first_'' to track when we're done
         // with, well, the first population.
         else
         {
            // If done_first_ is true, then we've iterated through all
            // of the first population and so are enumerating the
            // second.

            if ( done_first_ ) 
            {
               if ( current_ == second().end() )
               {
                  current_    = first().begin();

                                // reset this to indicate that we're
                                // back on the first population
                  done_first_ = false; 

               }
            }
            else
            {
               if ( current_ == first().end() )
               {
                  current_    = second().begin();

                                // we're done with the first pop and are 
                                // now iterating the second, so set
                                // this to true to indicate that

                  done_first_ = true; 

               }
            }
         }

         return *(current_++);

      }  // void operator()( Population::value_type & i ) 


      /// do any post-sequencing cleanup
      /**
         In this case, we reset the done_first_ variable to false so
         that we're sure to start with the first() population.
      */
      virtual void cleanup( Population & first )
      {
         done_first_ = false;
      }

      virtual void cleanup( Population & first, Population & second )
      {
         done_first_ = false;
      }

   private:

      /**
          set to \c true if we've finished iterating
          the first population; only relevent when
          we have enumerate two populations

          \note
          XXX but will we ever be using two pops?
       */
      bool done_first_;

      /// the current individual
      typename Population::iterator current_;

      /// initialize the selector
      /**
         Set the iterator to the first individual and set done_first_
         to \c false.
      */
      void init_( Population & first )
      {
         current_    = first.begin();
         done_first_ = false;
      }

      void init_( Population & first, Population & second )
      {
         current_    = first.begin();
         done_first_ = false;
      }

}; // class ecSequentialSelector






/// Arbitrarily select N individuals from the population and return one
/**
   \note 

   Insure that the populations total to more than two individuals,
   otherwise this will generate an infinite loop.

   \todo XXX Yes, but do we REALLY want to control this via template parameters?  */
template<class Population, class Comparator>
class ecTournamentSelector : public ecSelector<Population>
{
   public:

      /// Set the number of individuals to select from
      /**
         Defaults to two, or binary selection.
       */
      ecTournamentSelector( size_t selection_size = 2 )
         : selection_size_( selection_size )
      {}


      /// select an invididual
      virtual typename Population::value_type & get()
      {
         Comparator comparator;

         typename Population::value_type * a;
         typename Population::value_type * best = 
            &( uniform_selector_() );

         int selection_count = 2; // how many individuals we've selected
                                // so far
         do
         {
            a = &( uniform_selector_() );

                                // if the two objects are identical,
                                // then re-select until we get a
                                // different object; note this can
                                // lead to pathological behavior if
                                // there are only two individuals to
                                // choose from.
            if ( a == best )
            {
               selection_count--;
               continue;
            }

            if ( comparator( *a, *best ) )
            {
               best = a;
            }

            selection_count++;

         } while ( selection_count < selection_size_ );

         // and then return the "best" (or worse, depending on the Comparator)
         return *best;
      }


      /// do any post-processing cleanup
      virtual void cleanup( Population & first )
      {
         // NOP
      }
      virtual void cleanup( Population & first, Population & second )
      {
         // NOP
      }

   private:

      /// set up the uniform selector to select from the first population
      /**
         \todo XXX would this scenario ever occur?
       */
      void init_( Population & first )
      {
         uniform_selector_.init( first );
      }

      /// set up the uniform selector to select from both populations
      void init_( Population & first, Population & second )
      {
         uniform_selector_.init( first, second );
      }

      /// the number of individuals to select from; defaults to 2
      size_t selection_size_;

      /// uniform selector used to arbitrarily select individuals
      ecUniformSelector<Population> uniform_selector_;
}; // class ecTournamentSelector




#endif
