//
// ecStatistics.h
//

#ifndef EC_STATISTICS_H
#define EC_STATISTICS_H


/// helper function used to accumulate fitness values
/**
   Used in ecStatistics<Population>::update_local_information_().
 */
template<class Individual>
class 
accumulate_fitness_
{
   public:

      typename Individual::fitness_type operator()( typename Individual::fitness_type total,
                                           Individual const & i )
      {
         return total + i.fitness();
      }
}; // accumulate_fitness_



/// Holds statistics collected during an EA's run.
template<class Population>
class ecStatistics
{
   public:
      ecStatistics() 
         : global_average_( 0 ),
           local_average_( 0 ),
           local_standard_deviation_( 0.0 ),
           births_( 0 ),
           deaths_( 0 )
      {
      }

      ///
      virtual ~ecStatistics() {}

      /// return the number of births so far
      int         births() const     { return births_; }
      
      /** increment the birth count; we need the individual so that we
       can update the running global average
      */
      void        addBirth( typename Population::value_type const & );

      /// return the number of deaths that have occured so far
      int         deaths() const         { return deaths_; }

      /// add 'd' to the deaths count
      void        addDeaths( int d = 1 ) { deaths_ += d; }


      ///
      void
      ecStatistics::reset()
      {
         typename Population::value_type tmp_indiv = global_maximum_;

         global_maximum_ = global_minimum_;
         global_minimum_ = tmp_indiv;

         local_maximum_  = global_maximum_;
         local_minimum_  = global_minimum_;

         global_average_ = local_average_ = 0.0;

         global_average_ = 
            local_average_ = local_standard_deviation_ = 0.0;

         births_ = deaths_ = good_births_ = 0;

      } // ecStatistics::reset



      ///
      float
      globalAverage() const
      {
         return global_average_;
      } // globalAverage() const


      ///
      typename Population::value_type const &
      globalMaximum() const
      {
         return global_maximum_;
      } // globalMaximum() const


      ///
      typename Population::value_type const &
      globalMinimum() const
      {
         return global_minimum_;
      } // globalMinimum() const


      ///
      typename Population::value_type const &
      localMaximum() const
      {
         return local_maximum_;
      } // localMaximum() const


      ///
      typename Population::value_type const &
      localMinimum() const
      {
         return local_minimum_;
      } // localMinimum() const


      ///
      float
      localAverage() const
      {
         return local_average_;
      } // localAverage() const



      ///
      float
      localStandardDeviation() const
      {
         return local_standard_deviation_;
      } // localStandardDeviation() const


      ///
      /**!
       called by ecEA before the evolution loop; used for
       initializing variables
      */
      void operator()( Population const & parents )
      {
         update_local_information_( parents );

         global_maximum_ = local_maximum_;
         global_minimum_ = local_minimum_;

         // give the sub-classes a crack at the parents before
         // evolution begins, too
         update_( parents );

      }

      ///
      void operator()( Population const & parents, 
                       Population const & children )
      {
         update_local_information_( parents );

         births_ += children.size();    // naive assumption that the size of the
                                // children indicates the number of
                                // total births during the last generation

         // ... and update the global memory, if necessary

         global_maximum_ = max( local_maximum_, global_maximum_ );
         global_minimum_ = min( local_minimum_, global_minimum_ );


         // XXX need to add generation of other statistics like
         // XXX standard deviation and global average

         // give sub-classes a chance to do something during the update
         update_( parents, children );

      } // invoked within ecEA for tracking statistics for updating
      // maximums, minimums, averages, and standard deviations

   protected:

      ///
      virtual void update_( Population const & parents )
      {
         // NOP by default
      } // update_
      // this is optionally implemented in a sub-class for additional
      // updating and possibly printing information


      ///
      virtual void update_( Population const & parents,
                            Population const & children )
      {
         // NOP by default
      } // update_
      // this is optionally implemented in a sub-class for additional
      // updating and possibly printing information


   private:

      ///
      void update_local_information_( Population const & parents )
      {
         float local_fitnesses_ = 
            std::accumulate( parents.begin(),
                             parents.end(),
                             0.0,
                             accumulate_fitness_<typename Population::value_type>() );


         local_average_ = local_fitnesses_ / parents.size();

         total_fitnesses_ += local_fitnesses_;


         // remember the best and worst individuals for the current generation

         local_maximum_ = *(max_element( parents.begin(), parents.end() ) );
         local_minimum_ = *(min_element( parents.begin(), parents.end() ) );

      } // update_local_information_(); 


      /// total of all fitnesses so far; used to calculate global_average_
      float        total_fitnesses_;

      ///
      typename Population::value_type   global_maximum_;
      ///
      typename Population::value_type   global_minimum_;
      ///
      float                             global_average_;

      ///
      typename Population::value_type   local_maximum_;
      ///
      typename Population::value_type   local_minimum_;
      ///
      float                             local_average_;
      ///
      float                             local_standard_deviation_;

      /// all births, even aborted ones
      int          births_;
      

      /**! number of individuals killed when child and parent pops
       folded together 
      */
      int          deaths_;

}; // ecStatistics



#endif


