//
// ecStrategy.h
//

#ifndef EC_STRATEGY_H
#define EC_STRATEGY_H


/**!
 An ecStrategy is applies some specialized form of semantics onto a
 parent and/or child population.  There are roughly two families of
 strategy classes: those used for reproduction, and those used for
 survival.

 This somewhat employs the Strategy design pattern. _Design
 Patterns_, pgs. 315-323.
*/
template< class Population >
class  
ecStrategy
{
   public:

      virtual ~ecStrategy() {}

      //! Do any strategy specific initialization
      virtual void init( Population & parents,
                 Population & children ) 
      { 
         init_( parents, children ); 
      }


      //! This version is for operators that ignore parents.
      /**! XXX Is this too much a kludge? */
      virtual void init( Population & children ) 
      { 
         init_( children ); 
      }

      //!
      virtual void cleanup( Population & parents,
                            Population & children )
      { 
         cleanup_( parents, children );
      }


      //! This version is for operators that ignore parents.
      /**! XXX Is this too much a kludge? */
      virtual void cleanup( Population & children )
      { 
         cleanup_( children );
      }


   protected:

      /**!
       This is a calling entry point before the main body of the
       strategy, and will mostly be used by their corresponding
       operator to set up for work.  The parents and children may or
       may not be used by the particular strategy; they're passed in
       as it's better to not need them and have them than to need
       them and not have them.
      */
      virtual void init_( Population & parents,
                          Population & children ) 
      { /* NOP */ }


      // This version is for operators that ignore parents.
      // XXX Is this too much a kludge?
      virtual void init_( Population & children ) 
      { /* NOP */ }


      /**!
       This is a calling entry point after the main body of the
       strategy, and will mostly be used by their corresponding
       operators for post-processing.  (See previous comment on why
       we pass in parent and children containers.)
      */
      virtual void cleanup_( Population & parents,
                             Population & children )
      { /* NOP */ }


      //! This version is for operators that ignore parents.
      /**! XXX Is this too much a kludge? */
      virtual void cleanup_( Population & children )
      { /* NOP */ }

}; // ecStrategy



/**!
 This defines the kinds of strategies that work in tandem with some
 form of operator.  The reproduction strategies will typically use
 some form of birth operator; the survival strategies will use some
 kind of selection operator.
*/
template< class Population, class Operator >
class 
ecStrategyWithOperator : public ecStrategy<Population>
{
   public:

      virtual ~ecStrategyWithOperator() {}

      ecStrategyWithOperator( Operator & o )
         : operator_( o )
      {}

      //!
      Operator &       op()       { return operator_; }
      //!
      Operator const & op() const { return operator_; }

      //! Do any strategy specific initialization
      virtual void init( Population & parents,
                         Population & children ) 
      { 
         op().init( parents, children );

         init_( parents, children ); 
      }


      //! This version is for operators that ignore parents.
      /** XXX Is this too much a kludge? */
      virtual void init( Population & children ) 
      { 
         op().init( children );

         init_( children ); 
      }


      //!
      virtual void cleanup( Population & parents,
                            Population & children )
      { 
         op().cleanup( parents, children );

         cleanup_( parents, children );
      }


      //! This version is for operators that ignore parents.
      /** XXX Is this too much a kludge? */
      virtual void cleanup( Population & children )
      { 
         op().cleanup( children );
         cleanup_( children );
      }

   private:

      //! hidden because of reference member
      ecStrategyWithOperator() {}
      
      //!
      Operator & operator_;

}; // ecStrategyWithOperator


#endif
