//
// ecSurvive.h
//
// $Id: ecSurvive.h,v 1.8 2001/02/20 19:24:18 mcoletti Exp $
//

#ifndef EC_SURVIVE_H
#define EC_SURVIVE_H

#include <iostream>
#include <iterator>
#include <algorithm>
#include <queue>


#ifndef EC_OPERATOR_H
#include <eclib/ecOperator.h>
#endif


/**!
 Abstract base class for all survival operators which are, in turn,
 used by ecSurvivalStrategy objects to pare down one or two
 populations to a single population.  (I.e., (mu + lambda) vs. (mu ,
 lambda))
*/
template< class Population, 
          class SelectionOperator, 
          class Compare = std::less<typename Population::value_type> >
class ecSurvivalOperator : public ecDoubleOperator<Population>
{
   public:

      ecSurvivalOperator( SelectionOperator & so )
         : selection_operator_( so )
      {}

      virtual ~ecSurvivalOperator() {}

      //! Returns a new child.
      virtual typename Population::value_type operator()( ) = 0;

      //!
      SelectionOperator const & selector() const
      { return selection_operator_; }

      //!
      SelectionOperator & selector()
      { return selection_operator_; }

      //!
      virtual void init( Population & first )
      {
         selector().init( first );

         // now give the child class a chance at doing its own 
         // initialization
         init_( first );
      }


      //! XXX would this be used?  Does this make sense?
      virtual void init( Population & first, Population & second )
      {
         selector().init( first, second );

         // now give the child class a chance at doing its own 
         // initialization
         init_( first, second );
      }

      //!
      virtual void cleanup( Population & first )
      {
         // now give the child class a chance at doing its own 
         // cleanup
         cleanup_( first );
      }

      //!
      virtual void cleanup( Population & first, Population & second )
      {
         // now give the child class a chance at doing its own 
         // cleanup
         cleanup_( first, second );
      }


   protected:

      //! Child specific init called from cleanup()
      virtual void init_( Population & first )
      {
         // NOP, though child can obviously over-ride
      }

      //! Child specific init called from cleanup()
      virtual void init_( Population & first, Population & second )
      {
         // NOP, though child can obviously over-ride
      }

      //! Child specific cleanup called from cleanup()
      virtual void cleanup_( Population & first )
      {
         // NOP, though child can obviously over-ride
      }

      //! Child specific cleanup called from cleanup()
      virtual void cleanup_( Population & first, Population & second )
      {
         // NOP, though child can obviously over-ride
      }

   private:

      //! reference to operator used to select parents
      SelectionOperator & selection_operator_;

}; // ecSurvivalOperator



/**!
 A simple truncation survival algorithm.  Essentially a set of
 survivors is chosen from the best of the parents and children and
 copied into survivors.  The number of survivors is determined by
 the size of the survivors container.
*/
template< class Population, 
          class SelectionOperator,
          class Compare = std::less<typename Population::value_type>  >
class ecTruncationSurvival : public ecSurvivalOperator<Population,SelectionOperator,Compare>
{
   public:

      ecTruncationSurvival( SelectionOperator & so )
         : ecSurvivalOperator<Population,SelectionOperator,Compare>( so )
      {}


      virtual ~ecTruncationSurvival() {}


      //! Returns latest survivor
      /** XXX consider mitigating copy resource penalty */
      virtual typename Population::value_type  operator()( ) 
      {
         typename Population::value_type tmp_individual = survivors_.top();

         survivors_.pop();

         return tmp_individual;
      } // typename Population::value_type operator()( )



   protected:

      /**! XXX should we have a separate ``size'' value for truncation survival?
       XXX instead of using population sizes? 
      */
      void init_( Population  & first )
      {
         // slosh all the individuals from the population into the a
         // priority queue which implicitly sorts them from best to least
         // fit

         clear();               // start from clean slate

         for ( size_t i = 0;
               i < first.size();
               i++ )
         {
            survivors_.push( selector().get() );
         }

#ifdef DEBUG
         cout << survivors_.size() << " survivors\n";
#endif

      } // init_ for single population


      //!
      void init_( Population  & first,
                  Population  & second )
      {
         // slosh all the individuals from both populations into the a
         // priority queue which implicitly sorts them from best to least
         // fit

         clear();               // start from clean slate


//           cout << "first:\t" << first.size() << endl;
//           cout << "second:\t" << second.size() << endl;
//           cout << "survivors:\t" << survivors_.size() << endl;

         // XXX we presume that the given selector is smart enough to
         // XXX choose from both populations


         for ( size_t i = 0;
               i < first.size() + second.size();
               i++ )
         {
            survivors_.push( selector().get() );
         }


#ifdef DEBUG
         cout << survivors_.size() << " survivors\n";
#endif
      } // init_ for two populations



      //! Child specific cleanup called from cleanup()
      virtual void cleanup_( Population & first )
      {
         clear();
      }

      //! Child specific cleanup called from cleanup()
      virtual void cleanup_( Population & first, Population & second )
      {
         clear();
      }



   private:

      //! clear out any survivors
      void clear()
      {
         while ( survivors_.size() )
         {
            survivors_.pop();
         }
      }

      //! Contains the survivors from one or both populations
      std::priority_queue<typename Population::value_type, Population, Compare> 
         survivors_;
         

}; // class ecTruncationSurvival







template< class Population, class SelectionOperator >
class ecBinaryTournamentSurvivalOperator : public ecSurvivalOperator<Population,SelectionOperator>
{
   public:

      //! Returns a new child.
      /**
         Essentially take two individuals from the selector and return
         the one with the higher fitness.  

         \note 

         Naturally weird selectors would have odd behavior.  For
         example, using a sequential selector would make no sense.
         This doesn't check for that.  I.e., you'd better know what
         you're doing.j

      */
      virtual typename Population::value_type operator()( )
      {
         return std::max( selector().get(), selector().get() );
      }


   protected:

      //! Child specific init called from cleanup()
      virtual void init_( Population & first )
      {
         // we're happy with how the parent class attached the
         // selector to the population
      }

      //! Child specific init called from cleanup()
      virtual void init_( Population & first, Population & second )
      {
         // we're happy with how the parent class attached the
         // selector to the population
      }

      //! Child specific cleanup called from cleanup()
      virtual void cleanup_( Population & first )
      {
         // no house keeping need be done
      }

      //! Child specific cleanup called from cleanup()
      virtual void cleanup_( Population & first, Population & second )
      {
         // no house keeping need be done
      }

   private:

      // XXX I don't think this should be here.  JKB

      //! reference to operator used to select parents
      SelectionOperator & selection_operator_;

}; // ecSurvivalOperator




/**!
 A survival operator which allows everything the selector picks to survive.
 This is useful for EAs which do not implement survival selection.  The
 canonical GA is a perfect example of this.  A GA can be thought of as a
 (mu,lambda) algorithm where mu = lambda, and the survival selection operator
 is sequential.

 Actually, it should be possible to implement tournament survival selection
 using this survival operator by just handing it an ecTournamentSelector.
 Other useful combinations may be possible too.
*/
template< class Population, class SelectionOperator >
class ecNullSurvivalOperator : public ecSurvivalOperator<Population,SelectionOperator>
{
   public:

      ecNullSurvivalOperator( SelectionOperator & so )
         : ecSurvivalOperator<Population,SelectionOperator>( so )
      {}

      //! Returns a survivor.
      virtual typename Population::value_type operator()( )
      {
          return selector().get();
      }


   protected:

      //! Child specific init called from cleanup()
      virtual void init_( Population & first )
      {
         // we're happy with how the parent class attached the
         // selector to the population
      }

      //! Child specific init called from cleanup()
      virtual void init_( Population & first, Population & second )
      {
         // we're happy with how the parent class attached the
         // selector to the population
      }

      //! Child specific cleanup called from cleanup()
      virtual void cleanup_( Population & first )
      {
         // no house keeping need be done
      }

      //! Child specific cleanup called from cleanup()
      virtual void cleanup_( Population & first, Population & second )
      {
         // no house keeping need be done
      }

}; // ecNullSurvivalOperator






#endif
