//
// ecEA_t.cpp
//
// $Id: ecEA_t.cpp,v 1.5 2000/10/18 19:13:05 mcoletti Exp $
//


#include <algorithm>
#include <iterator>
#include <functional>
#include <numeric>
#include <iostream>
#include <cassert>


using namespace std;


#include <eclib/ecIndividual.h>
#include <eclib/ecPopulation.h>
#include <eclib/ecRandom.h>
#include <eclib/ecBirth.h>
#include <eclib/ecSelector.h>
#include <eclib/ecHalt.h>
#include <eclib/ecReproStrategy.h>
#include <eclib/ecSurvive.h>
#include <eclib/ecSurvivalStrategy.h>
#include <eclib/ecStatistics.h>
#include <eclib/ecEA.h>


// A test individual class for the x^2 landscape
//
// Used to exercise ecIndividual oriented tests only.  More
// sophisticated tests are in other test programs.
//
class SimpleIndividual : public ecIndividual< float, float >
{
   public:

      SimpleIndividual()
         : ecIndividual< float, float >( 2 )
      {}
      // Individuals of this type will have two elements in their genome

      void randomize()
      {
         // randomize the genes to values in [0,4]
         for ( iterator i = begin(); i != end(); i++ )
         {
            *i = drand() * 4;
         }
      }

      float evaluate() const
      {
         return accumulate( begin(), end(), 1.0, multiplies<float>() );
      }


   protected:

      void print_( ostream & os ) const
      {
         // print the birthday, followed by the gene values, then the fitness

         os << "( ";

         ostream_iterator<float> ostr_itr( os, " " );
         copy( begin(), end(), ostr_itr );

         os << "), "
            << fitness() 
            << "\n";
      }

}; // SimpleIndividual


typedef ecPopulation<SimpleIndividual> Population;
typedef ecGenerationHalt<Population>   StopAtGeneration;


int
main( int argc, char** argv )
{
   Population parents;

   // let's add some guys to the parents
   for ( int i = 0; i < 10; i++ )
   {
      parents.push_back( SimpleIndividual() );
   }

   ostream_iterator<SimpleIndividual> ostr_itr( cout, "\n" );

   // give them initial random values
   randomize( parents.begin(), parents.end() );

   cout << "parents:\n";
   copy( parents.begin(), parents.end(), ostr_itr );


   // the next three steps necessary to make a create function to be
   // passed into the EA may indicate a design problem -- a lot of
   // hand waving here

   typedef ecGaussianMutationBirthOperator< ecPopulation<SimpleIndividual>, 
      ecSequentialSelector< ecPopulation<SimpleIndividual> > > GaussianBOP;

   typedef ecSequentialSelector< ecPopulation<SimpleIndividual> > SequentialSelector;


   // first build the reproduction strategy components, and then the
   // reproductive strategy itself

   SequentialSelector birth_sequential_selector;


   GaussianBOP gaussian_bop( 
      birth_sequential_selector,
      0.5, 
      0.75 );

   gaussian_bop.init( parents );

   ecReproductionStrategy< Population, GaussianBOP > gaussianCreate( gaussian_bop );


   // next, build the surival strategy components followed by the
   // survival strategy

   SequentialSelector survival_sequential_selector;

   ecTruncationSurvival< Population, SequentialSelector > 
      truncationSurvival( survival_sequential_selector );

   ecMuPlusLambdaSurvivalStrategy<Population, 
                                 ecTruncationSurvival<Population,SequentialSelector> > 
      mySurvivalStrategy( truncationSurvival );


   StopAtGeneration stopAtGeneration( 30 );

   ecStatistics<Population> statistics;

   // invoke a simple ES style algorithm
   ecEA< Population, 
         StopAtGeneration, 
         ecReproductionStrategy< Population, GaussianBOP > ,
         ecMuPlusLambdaSurvivalStrategy<Population, 
             ecTruncationSurvival<Population,SequentialSelector> > ,
         ecStatistics<Population>
       >
      ( parents,
        20,
        stopAtGeneration,
        gaussianCreate,
        mySurvivalStrategy,
        statistics
      );
      
   cout << "parents after 30 generations:\n";

   copy( parents.begin(), parents.end(), ostr_itr );

   exit( 0 );
}



