//
// ecSelector_t.cpp
//
// $Id: ecSelector_t.cpp,v 1.5 2000/10/18 19:13:05 mcoletti Exp $
//


#include <algorithm>
#include <functional>
#include <numeric>
#include <iostream>
#include <cassert>


using namespace std;


#include <eclib/ecIndividual.h>
#include <eclib/ecPopulation.h>
#include <eclib/ecRandom.h>
#include <eclib/ecBirth.h>
#include <eclib/ecSelector.h>


// A test individual class for the x^2 landscape
//
// Used to exercise ecIndividual oriented tests only.  More
// sophisticated tests are in other test programs.
//
class SimpleIndividual : public ecIndividual< float, float >
{
   public:

      SimpleIndividual()
         : ecIndividual< float, float >( 2 )
      {}
      // Individuals of this type will have two elements in their genome

      void randomize()
      {
         // randomize the genes to values in [0,4]
         for ( iterator i = begin(); i != end(); i++ )
         {
            *i = drand() * 4;
         }
      }

      float evaluate() const
      {
         return accumulate( begin(), end(), 1.0, multiplies<float>() );
      }


   protected:

      void print_( ostream & os ) const
      {
         os << ", ( ";

         ostream_iterator<float> ostr_itr( os, " " );
         copy( begin(), end(), ostr_itr );

         os << "), "
            << fitness() 
            << "\n";
      }

}; // SimpleIndividual


struct MakeIndividual
{
      SimpleIndividual operator()()
      {
         return SimpleIndividual();
      }
}; // struct MakeIndividual



typedef ecPopulation<SimpleIndividual> Population;



int
main( int argc, char** argv )
{
   // create a simple population of ten individuals first

   Population population;

   generate_n( back_inserter(population), 10, MakeIndividual() );


   // blat 'em out

   cout << "popluation before randomization:\n";

   ostream_iterator<SimpleIndividual> ostr_itr( cout, "\n" );
   copy( population.begin(), population.end(), ostr_itr );


   // stir 'em up a bit
   randomize( population.begin(), population.end() );

   cout << "popluation after randomization:\n";

   copy( population.begin(), population.end(), ostr_itr );

   {
      // now create a new population using uniform selection

      Population new_population;

      ecUniformSelector<Population> selector;

      selector.init(population);

      generate_n( back_inserter(new_population),
                  population.size(),
                  selector );

      cout << "new population with individuals randomly selected from the first:\n";

      copy( new_population.begin(), new_population.end(), ostr_itr );

   }



   {
      // now create a new population using uniform selection from TWO
      // populations

      Population new_population, second_population;

      // now create the second population with ten individuals of all zeroes
      generate_n( back_inserter(second_population), 10, MakeIndividual() );


      ecUniformSelector<Population> selector;

      selector.init( population, second_population );

      generate_n( back_inserter(new_population),
                  15,
                  selector );

      cout << "new population with individuals randomly selected from the first and second populations:\n";

      copy( new_population.begin(), new_population.end(), ostr_itr );

   }


   {
      // now create a new population using sequential selection

      Population new_population;

      ecSequentialSelector< Population > selector;
      selector.init(population);

      generate_n( back_inserter(new_population),
                  population.size(),
                  selector );

      cout << "new population with individuals sequentially selected from the first:\n";

      copy( new_population.begin(), new_population.end(), ostr_itr );

   }




   {
      // now create a new population using sequential selection from TWO
      // populations

      Population new_population, second_population;

      // now create the second population with ten individuals of all zeroes
      generate_n( back_inserter(second_population), 10, MakeIndividual() );


      ecSequentialSelector<Population> selector;

      selector.init( population, second_population );

      generate_n( back_inserter(new_population),
                  15,
                  selector );

      cout << "new population with individuals sequentially selected from the first and second populations:\n";

      copy( new_population.begin(), new_population.end(), ostr_itr );

   }




   {
      // now create a new population using binary tournament selection

      Population new_population, second_population;

      ecTournamentSelector< Population, std::less<Population::value_type> >
         selector;

      // now create the second population with five individuals of all zeroes
      generate_n( back_inserter(second_population), 5, MakeIndividual() );

      cout << "population:\n";
      copy( population.begin(), population.end(), ostr_itr );

      cout << "second_population:\n";
      copy( second_population.begin(), second_population.end(), ostr_itr );

      selector.init( population, second_population );

      // Generate enough new individuals such that we wrap back to the
      // first population.  That way we test not only jumping to the
      // second population when the first is enumerated, but wrapping
      // back to the first population when the second is completely
      // enumerated.
      generate_n( back_inserter(new_population),
                  population.size() ,
                  selector );

      cout << "new population with individuals selected via binary tournament from the first and second:\n";

      copy( new_population.begin(), new_population.end(), ostr_itr );

      // now use more<> not less<>


      ecTournamentSelector< Population, std::greater<Population::value_type> >
         new_selector;

      new_selector.init( population, second_population );

      new_population.clear();

      generate_n( back_inserter(new_population),
                  population.size() ,
                  new_selector );

      cout << "new population with individuals selected via binary tournament from the first and second with \"greater than bias\":\n";

      copy( new_population.begin(), new_population.end(), ostr_itr );


   }

   exit( 0 );

}

