//
// ecSurvive_t.cpp
//
// $Id: ecSurvive_t.cpp,v 1.6 2001/02/20 19:24:18 mcoletti Exp $
//


#include <algorithm>
#include <iterator>
#include <functional>
#include <numeric>
#include <iostream>
#include <cassert>


using namespace std;


#include <eclib/ecIndividual.h>
#include <eclib/ecPopulation.h>
#include <eclib/ecRandom.h>
#include <eclib/ecBirth.h>
#include <eclib/ecSelector.h>
#include <eclib/ecSurvive.h>


// A test individual class for the x^2 landscape
//
// Used to exercise ecIndividual oriented tests only.  More
// sophisticated tests are in other test programs.
//
class SimpleIndividual : public ecIndividual< float, float >
{
   public:

      SimpleIndividual()
         : ecIndividual< float, float >( 2 )
      {}
      // Individuals of this type will have two elements in their genome

      void randomize()
      {
         // randomize the genes to values in [0,4]
         for ( iterator i = begin(); i != end(); i++ )
         {
            *i = drand() * 4;
         }
      }

      float evaluate() const
      {
         return accumulate( begin(), end(), 1.0, multiplies<float>() );
      }


   protected:

      void print_( ostream & os ) const
      {
         os << "( ";

         ostream_iterator<float> ostr_itr( os, " " );
         copy( begin(), end(), ostr_itr );

         os << "), "
            << fitness() 
            << "\n";
      }

}; // SimpleIndividual



typedef ecPopulation<SimpleIndividual> Population;


// Create an initial parent population.  Then create a child
// population.  We arbitrarily choose the gauassian mutation birth
// operator to create the child population.
//
void
init( size_t num_parents, 
      size_t num_children,
      Population & parents, 
      Population & children )
{

   // let's add some guys to the parents
   for ( int i = 0; i < num_parents; i++ )
   {
      parents.push_back( SimpleIndividual() );
   }

   ostream_iterator<SimpleIndividual> ostr_itr( cout, "\n" );

   // ... and stir them up a bit

   cout << "initial parents:\n";

   randomize( parents.begin(), parents.end() );

   copy( parents.begin(), parents.end(), ostr_itr );


   // now cut loose the gaussian mutation operator to create a child
   // population
   
   cout << "\n... and a child population generated via gaussian mutation:\n";

   
   ecSequentialSelector< Population > selector;
   selector.init( parents );

   generate_n( back_inserter( children ),
               num_children,
               ecGaussianMutationBirthOperator< Population, ecSequentialSelector< Population > >( selector, 0.5, 0.75 ) );


   copy( children.begin(), children.end(), ostr_itr );

} // init()



typedef ecSequentialSelector<Population> SequentialSelector;



int
main( int argc, char** argv )
{
   // we want to test out the truncation survival; first with just
   // one population, then with two

   {
      cout << "single population survival:\n\n";

      Population parents, children;

   // create ten parents and then twenty children from those parents

      init( 10, 20, parents, children );




      SequentialSelector selector;
      selector.init( children );


      ecTruncationSurvival< Population, SequentialSelector > 
         truncationSurvival( selector );

      truncationSurvival.init( children );


      Population survivors( 10 );   // have just ten survivors

      for ( Population::iterator i = survivors.begin();
            i != survivors.end();
            i++ )
      {
         *i = truncationSurvival();
      }


      ostream_iterator<SimpleIndividual> ostr_itr( cout, "\n" );

      cout << "\nThe survivors:\n";

      copy( survivors.begin(), survivors.end(), ostr_itr );

   }


   {
      cout << "double population survival:\n\n";


      Population parents, children;

      // create ten parents and then twenty children from those parents

      init( 10, 20, parents, children );




      SequentialSelector selector;

      ecTruncationSurvival< Population, SequentialSelector > 
         truncationSurvival( selector );

      truncationSurvival.init( parents, children );


      Population survivors( 10 );   // have just ten survivors

      for ( Population::iterator i = survivors.begin();
            i != survivors.end();
            i++ )
      {
         *i = truncationSurvival();
      }


      ostream_iterator<SimpleIndividual> ostr_itr( cout, "\n" );

      cout << "\nThe survivors:\n";

      copy( survivors.begin(), survivors.end(), ostr_itr );

   }

   // now we want to see if we can use a different comparator 
   {
      cout << "single population survival with std::greater:\n\n";

      Population parents, children;

   // create ten parents and then twenty children from those parents

      init( 10, 20, parents, children );




      SequentialSelector selector;
      selector.init( children );


      ecTruncationSurvival< Population, SequentialSelector, std::greater<Population::value_type>  > 
         truncationSurvival( selector );

      truncationSurvival.init( children );


      Population survivors( 10 );   // have just ten survivors

      for ( Population::iterator i = survivors.begin();
            i != survivors.end();
            i++ )
      {
         *i = truncationSurvival();
      }


      ostream_iterator<SimpleIndividual> ostr_itr( cout, "\n" );

      cout << "\nThe survivors:\n";

      copy( survivors.begin(), survivors.end(), ostr_itr );

   }


   {
      cout << "double population survival std::greater:\n\n";


      Population parents, children;

      // create ten parents and then twenty children from those parents

      init( 10, 20, parents, children );




      SequentialSelector selector;

      ecTruncationSurvival< Population, SequentialSelector, std::greater<Population::value_type> > 
         truncationSurvival( selector );

      truncationSurvival.init( parents, children );


      Population survivors( 10 );   // have just ten survivors

      for ( Population::iterator i = survivors.begin();
            i != survivors.end();
            i++ )
      {
         *i = truncationSurvival();
      }


      ostream_iterator<SimpleIndividual> ostr_itr( cout, "\n" );

      cout << "\nThe survivors:\n";

      copy( survivors.begin(), survivors.end(), ostr_itr );

   }



   // we want to test out the null survival operator; the survivors
   // should be identical to the children

   {
      cout << "null survival operator:\n\n";

      Population parents, children;

      // create ten parents and then twenty children from those parents

      init( 10, 10, parents, children );




      SequentialSelector selector;
      selector.init( children );


      ecNullSurvivalOperator< Population, SequentialSelector > 
         nullSurvival( selector );

      nullSurvival.init( children );


      Population survivors( 10 );   // have just ten survivors

      for ( Population::iterator i = survivors.begin();
            i != survivors.end();
            i++ )
      {
         *i = nullSurvival();
      }


      ostream_iterator<SimpleIndividual> ostr_itr( cout, "\n" );

      cout << "\nThe survivors:\n";

      copy( survivors.begin(), survivors.end(), ostr_itr );

   }


   exit( 0 );
}

